# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>
#            Matt Prahl <mprahl@redhat.com>
#            Jan Kaluza <jkaluza@redhat.com>
from module_build_service import log, models, Modulemd, conf, db
from module_build_service.errors import StreamAmbigous
from module_build_service.mmd_resolver import MMDResolver
from module_build_service import glib
import module_build_service.resolver


def _expand_mse_streams(session, name, streams, default_streams, raise_if_stream_ambigous):
    """
    Helper method for `expand_mse_stream()` expanding single name:[streams].
    Returns list of expanded streams.

    :param session: SQLAlchemy DB session.
    :param str name: Name of the module which will be expanded.
    :param list streams: List of streams to expand.
    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    """
    default_streams = default_streams or {}
    # Stream can be prefixed with '-' sign to define that this stream should
    # not appear in a resulting list of streams. There can be two situations:
    # a) all streams have '-' prefix. In this case, we treat list of streams
    #    as blacklist and we find all the valid streams and just remove those with
    #    '-' prefix.
    # b) there is at least one stream without '-' prefix. In this case, we can
    #    ignore all the streams with '-' prefix and just add those without
    #    '-' prefix to the list of valid streams.
    streams_is_blacklist = all(stream.startswith("-") for stream in streams.get())
    if streams_is_blacklist or len(streams.get()) == 0:
        if name in default_streams:
            expanded_streams = [default_streams[name]]
        elif raise_if_stream_ambigous:
            raise StreamAmbigous(
                "There are multiple streams to choose from for module %s." % name)
        else:
            builds = models.ModuleBuild.get_last_build_in_all_streams(
                session, name)
            expanded_streams = [build.stream for build in builds]
    else:
        expanded_streams = []
    for stream in streams.get():
        if stream.startswith("-"):
            if streams_is_blacklist and stream[1:] in expanded_streams:
                expanded_streams.remove(stream[1:])
        else:
            expanded_streams.append(stream)

    if len(expanded_streams) > 1:
        if name in default_streams:
            expanded_streams = [default_streams[name]]
        elif raise_if_stream_ambigous:
            raise StreamAmbigous("There are multiple streams %r to choose from for module %s."
                                 % (expanded_streams, name))

    return expanded_streams


def expand_mse_streams(session, mmd, default_streams=None, raise_if_stream_ambigous=False):
    """
    Expands streams in both buildrequires/requires sections of MMD.

    :param session: SQLAlchemy DB session.
    :param Modulemd.Module mmd: Modulemd metadata with original unexpanded module.
    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    """
    for deps in mmd.get_dependencies():
        expanded = {}
        for name, streams in deps.get_requires().items():
            streams_set = Modulemd.SimpleSet()
            streams_set.set(_expand_mse_streams(
                session, name, streams, default_streams, raise_if_stream_ambigous))
            expanded[name] = streams_set
        deps.set_requires(expanded)

        expanded = {}
        for name, streams in deps.get_buildrequires().items():
            streams_set = Modulemd.SimpleSet()
            streams_set.set(_expand_mse_streams(
                session, name, streams, default_streams, raise_if_stream_ambigous))
            expanded[name] = streams_set
        deps.set_buildrequires(expanded)


def _get_mmds_from_requires(session, requires, mmds, recursive=False,
                            default_streams=None, raise_if_stream_ambigous=False):
    """
    Helper method for get_mmds_required_by_module_recursively returning
    the list of module metadata objects defined by `requires` dict.

    :param session: SQLAlchemy DB session.
    :param requires: Modulemetadata requires or buildrequires.
    :param mmds: Dictionary with already handled name:streams as a keys and lists
        of resulting mmds as values.
    :param recursive: If True, the requires are checked recursively.
    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    :return: Dict with name:stream as a key and list with mmds as value.
    """
    default_streams = default_streams or {}
    # To be able to call itself recursively, we need to store list of mmds
    # we have added to global mmds list in this particular call.
    added_mmds = {}
    resolver = module_build_service.resolver.GenericResolver.create(conf)

    for name, streams in requires.items():
        streams_to_try = streams.get()
        if name in default_streams:
            streams_to_try = [default_streams[name]]
        elif len(streams_to_try) > 1 and raise_if_stream_ambigous:
            raise StreamAmbigous("There are multiple streams %r to choose from for module %s."
                                 % (streams_to_try, name))

        # For each valid stream, find the last build in a stream and also all
        # its contexts and add mmds of these builds to `mmds` and `added_mmds`.
        # Of course only do that if we have not done that already in some
        # previous call of this method.
        for stream in streams.get():
            ns = "%s:%s" % (name, stream)
            if ns in mmds:
                continue

            mmds[ns] = resolver.get_module_modulemds(name, stream, strict=True)
            added_mmds[ns] = mmds[ns]

    # Get the requires recursively.
    if recursive:
        for mmd_list in added_mmds.values():
            for mmd in mmd_list:
                for deps in mmd.get_dependencies():
                    mmds = _get_mmds_from_requires(session, deps.get_requires(), mmds, True)

    return mmds


def get_mmds_required_by_module_recursively(
        session, mmd, default_streams=None, raise_if_stream_ambigous=False):
    """
    Returns the list of Module metadata objects of all modules required while
    building the module defined by `mmd` module metadata. This presumes the
    module metadata streams are expanded using `expand_mse_streams(...)`
    method.

    This method finds out latest versions of all the build-requires of
    the `mmd` module and then also all contexts of these latest versions.

    For each build-required name:stream:version:context module, it checks
    recursively all the "requires" and finds the latest version of each
    required module and also all contexts of these latest versions.

    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    :rtype: list of Modulemd metadata
    :return: List of all modulemd metadata of all modules required to build
        the module `mmd`.
    """
    # We use dict with name:stream as a key and list with mmds as value.
    # That way, we can ensure we won't have any duplicate mmds in a resulting
    # list and we also don't waste resources on getting the modules we already
    # handled from DB.
    mmds = {}

    # At first get all the buildrequires of the module of interest.
    for deps in mmd.get_dependencies():
        mmds = _get_mmds_from_requires(
            session, deps.get_buildrequires(), mmds, False, default_streams,
            raise_if_stream_ambigous)

    # Now get the requires of buildrequires recursively.
    for mmd_key in list(mmds.keys()):
        for mmd in mmds[mmd_key]:
            for deps in mmd.get_dependencies():
                mmds = _get_mmds_from_requires(
                    session, deps.get_requires(), mmds, True, default_streams,
                    raise_if_stream_ambigous)

    # Make single list from dict of lists.
    res = []
    for mmds_list in mmds.values():
        res += mmds_list
    return res


def generate_expanded_mmds(session, mmd, raise_if_stream_ambigous=False, default_streams=None):
    """
    Returns list with MMDs with buildrequires and requires set according
    to module stream expansion rules. These module metadata can be directly
    built using MBS.

    :param session: SQLAlchemy DB session.
    :param Modulemd.Module mmd: Modulemd metadata with original unexpanded module.
    :param bool raise_if_stream_ambigous: When True, raises a StreamAmbigous exception in case
        there are multiple streams for some dependency of module and the module name is not
        defined in `default_streams`, so it is not clear which stream should be used.
    :param dict default_streams: Dict in {module_name: module_stream, ...} format defining
        the default stream to choose for module in case when there are multiple streams to
        choose from.
    """
    if not session:
        session = db.session

    if not default_streams:
        default_streams = {}

    # Create local copy of mmd, because we will expand its dependencies,
    # which would change the module.
    # TODO: Use copy method once its in released libmodulemd:
    # https://github.com/fedora-modularity/libmodulemd/pull/20
    current_mmd = Modulemd.Module.new_from_string(mmd.dumps())

    # MMDResolver expects the input MMD to have no context.
    current_mmd.set_context(None)

    # Expands the MSE streams. This mainly handles '-' prefix in MSE streams.
    expand_mse_streams(session, current_mmd, default_streams, raise_if_stream_ambigous)

    # Get the list of all MMDs which this module can be possibly built against
    # and add them to MMDResolver.
    mmd_resolver = MMDResolver()
    mmds_for_resolving = get_mmds_required_by_module_recursively(
        session, current_mmd, default_streams, raise_if_stream_ambigous)
    for m in mmds_for_resolving:
        mmd_resolver.add_modules(m)

    # Show log.info message with the NSVCs we have added to mmd_resolver.
    nsvcs_to_solve = [
        ":".join([m.get_name(), m.get_stream(), str(m.get_version()), str(m.get_context())])
        for m in mmds_for_resolving]
    log.info("Starting resolving with following input modules: %r", nsvcs_to_solve)

    # Resolve the dependencies between modules and get the list of all valid
    # combinations in which we can build this module.
    requires_combinations = mmd_resolver.solve(current_mmd)
    log.info("Resolving done, possible requires: %r", requires_combinations)

    # This is where we are going to store the generated MMDs.
    mmds = []
    for requires in requires_combinations:
        # Each generated MMD must be new Module object...
        # TODO: Use copy method once its in released libmodulemd:
        # https://github.com/fedora-modularity/libmodulemd/pull/20
        mmd_copy = Modulemd.Module.new_from_string(mmd.dumps())
        xmd = glib.from_variant_dict(mmd_copy.get_xmd())

        # Requires contain the NSVC representing the input mmd.
        # The 'context' of this NSVC defines the id of buildrequires/requires
        # pair in the mmd.get_dependencies().
        dependencies_id = None

        # We don't want to depend on ourselves, so store the NSVC of the current_mmd
        # to be able to ignore it later.
        self_nsvca = None

        # Dict to store name:stream pairs from nsvca, so we are able to access it
        # easily later.
        req_name_stream = {}

        # Get the values for dependencies_id, self_nsvca and req_name_stream variables.
        for nsvca in requires:
            req_name, req_stream, _, req_context, req_arch = nsvca.split(":")
            if req_arch == 'src':
                assert req_name == current_mmd.get_name()
                assert req_stream == current_mmd.get_stream()
                assert dependencies_id is None
                assert self_nsvca is None
                dependencies_id = int(req_context)
                self_nsvca = nsvca
                continue
            req_name_stream[req_name] = req_stream
        if dependencies_id is None or self_nsvca is None:
            raise RuntimeError(
                "%s:%s not found in requires %r" % (
                    current_mmd.get_name(), current_mmd.get_stream(), requires))

        # The name:[streams, ...] pairs do not have to be the same in both
        # buildrequires/requires. In case they are the same, we replace the streams
        # in requires section with a single stream against which we will build this MMD.
        # In case they are not the same, we have to keep the streams as they are in requires
        # section.  We always replace stream(s) for build-requirement with the one we
        # will build this MMD against.
        new_dep = Modulemd.Dependencies()
        dep = mmd_copy.get_dependencies()[dependencies_id]
        dep_requires = dep.get_requires()
        dep_buildrequires = dep.get_buildrequires()
        for req_name, req_streams in dep_requires.items():
            if req_name not in dep_buildrequires:
                # This require is not a buildrequire so just copy this runtime requirement to
                # new_dep and don't touch buildrequires
                new_dep.add_requires(req_name, req_streams.get())
            elif set(req_streams.get()) != set(dep_buildrequires[req_name].get()):
                # Streams in runtime section are not the same as in buildtime section,
                # so just copy this runtime requirement to new_dep.
                new_dep.add_requires(req_name, req_streams.get())
                new_dep.add_buildrequires(req_name, [req_name_stream[req_name]])
            else:
                # This runtime requirement has the same streams in both runtime/buildtime
                # requires sections, so replace streams in both sections by the one we
                # really used in this resolved variant.
                new_dep.add_requires(req_name, [req_name_stream[req_name]])
                new_dep.add_buildrequires(req_name, [req_name_stream[req_name]])
        mmd_copy.set_dependencies((new_dep, ))

        # The Modulemd.Dependencies() stores only streams, but to really build this
        # module, we need NSVC of buildrequires, so we have to store this data in XMD.
        # We also need additional data like for example list of filtered_rpms. We will
        # get them using module_build_service.resolver.GenericResolver.resolve_requires,
        # so prepare list with NSVCs of buildrequires as an input for this method.
        br_list = []
        for nsvca in requires:
            if nsvca == self_nsvca:
                continue
            # Remove the arch from nsvca
            nsvc = ":".join(nsvca.split(":")[:-1])
            br_list.append(nsvc)

        # Resolve the buildrequires and store the result in XMD.
        if 'mbs' not in xmd:
            xmd['mbs'] = {}
        resolver = module_build_service.resolver.GenericResolver.create(conf)
        xmd['mbs']['buildrequires'] = resolver.resolve_requires(br_list)
        xmd['mbs']['mse'] = True

        mmd_copy.set_xmd(glib.dict_values(xmd))

        # Now we have all the info to actually compute context of this module.
        ref_build_context, build_context, runtime_context, context = \
            models.ModuleBuild.contexts_from_mmd(mmd_copy.dumps())
        mmd_copy.set_context(context)

        mmds.append(mmd_copy)

    return mmds



# -*- coding: utf-8 -*-
# Copyright (c) 2017  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluza <jkaluza@redhat.com>

""" Handlers for repo change events on the message bus. """

import module_build_service.builder
import logging
import koji
from module_build_service import models, log, messaging

logging.basicConfig(level=logging.DEBUG)


def tagged(config, session, msg):
    """ Called whenever koji tags a build to tag. """
    if config.system not in ("koji", "test"):
        return []

    # Find our ModuleBuild associated with this tagged artifact.
    tag = msg.tag
    module_build = models.ModuleBuild.from_tag_change_event(session, msg)
    if not module_build:
        log.debug("No module build found associated with koji tag %r" % tag)
        return

    # Find tagged component.
    component = models.ComponentBuild.from_component_name(
        session, msg.artifact, module_build.id)
    if not component:
        log.error("No component %s in module %r", msg.artifact, module_build)
        return

    log.info("Saw relevant component tag of %r from %r." % (component.nvr,
                                                            msg.msg_id))

    # Mark the component as tagged
    if tag.endswith('-build'):
        component.tagged = True
    else:
        component.tagged_in_final = True
    session.commit()

    unbuilt_components_in_batch = [
        c for c in module_build.current_batch()
        if c.state == koji.BUILD_STATES['BUILDING'] or not c.state
    ]
    if unbuilt_components_in_batch:
        log.info("Not regenerating repo for tag %s, there are still "
                 "building components in a batch", tag)
        return []

    # Get the list of untagged components in current/previous batches which
    # have been built successfully.
    untagged_components = [
        c for c in module_build.up_to_current_batch()
        if (not c.tagged or (not c.tagged_in_final and not c.build_time_only)) and
        c.state == koji.BUILD_STATES['COMPLETE']
    ]

    further_work = []

    # If all components are tagged, start newRepo task.
    if not untagged_components:
        builder = module_build_service.builder.GenericBuilder.create_from_module(
            session, module_build, config)

        unbuilt_components = [
            c for c in module_build.component_builds
            if c.state == koji.BUILD_STATES['BUILDING'] or not c.state
        ]
        if unbuilt_components:
            repo_tag = builder.module_build_tag['name']
            log.info("All components in batch tagged, regenerating repo for tag %s", repo_tag)
            task_id = builder.koji_session.newRepo(repo_tag)
            module_build.new_repo_task_id = task_id
        else:
            # In case this is the last batch, we do not need to regenerate the
            # buildroot, because we will not build anything else in it. It
            # would be useless to wait for a repository we will not use anyway.
            log.info("All components in module tagged and built, skipping the "
                     "last repo regeneration")
            further_work += [messaging.KojiRepoChange(
                'components::_finalize: fake msg',
                builder.module_build_tag['name'])]
        session.commit()

    return further_work



config = dict(
    logging=dict(
        loggers=dict(
            # Quiet this guy down...
            requests={
                "level": "WARNING",
                "propagate": True,
                "handlers": ["console"],
            },
            module_build_service={
                "level": "INFO",
                "propagate": True,
                "handlers": ["console"],
            },
        ),
    ),
)



# -*- coding: utf-8 -*-
#
# Copyright © 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Jan Kaluža <jkaluza@redhat.com>
#            Igor Gnatenko <ignatenko@redhat.com>

import enum
import collections
import itertools
import solv
from module_build_service import log


class MMDResolverPolicy(enum.Enum):
    All = "all"      # All possible top-level combinations
    First = "first"  # All possible top-level combinations (filtered by N:S, first picked)


class MMDResolver(object):
    """
    Resolves dependencies between Module metadata objects.
    """

    def __init__(self):
        self.pool = solv.Pool()
        self.pool.setarch("x86_64")
        self.build_repo = self.pool.add_repo("build")
        self.available_repo = self.pool.add_repo("available")

    def _deps2reqs(self, deps):
        pool = self.pool

        rel_or_dep = lambda dep, op, rel: dep.Rel(op, rel) if dep is not None else rel
        stream_dep = lambda n, s: pool.Dep("module(%s:%s)" % (n, s))

        reqs = None
        for deps in deps:
            require = None
            for name, streams in deps.items():
                req_pos = req_neg = None
                for stream in streams:
                    if stream.startswith("-"):
                        req_neg = rel_or_dep(req_neg, solv.REL_OR, stream_dep(name, stream[1:]))
                    else:
                        req_pos = rel_or_dep(req_pos, solv.REL_OR, stream_dep(name, stream))

                req = pool.Dep("module(%s)" % name)
                if req_pos is not None:
                    req = req.Rel(solv.REL_WITH, req_pos)
                elif req_neg is not None:
                    req = req.Rel(solv.REL_WITHOUT, req_neg)

                require = rel_or_dep(require, solv.REL_AND, req)

            reqs = rel_or_dep(reqs, solv.REL_OR, require)

        return reqs

    def add_modules(self, mmd):
        n, s, v, c = mmd.get_name(), mmd.get_stream(), mmd.get_version(), mmd.get_context()

        pool = self.pool

        normdeps = lambda mmd, fn: [{name: streams.get()
                                     for name, streams in getattr(dep, fn)().items()}
                                    for dep in mmd.get_dependencies()]

        solvables = []
        if c is not None:
            # Built module

            # $n:$s:$v:$c-$v.$a
            solvable = self.available_repo.add_solvable()
            solvable.name = "%s:%s:%d:%s" % (n, s, v, c)
            solvable.evr = str(v)
            # TODO: replace with real arch
            solvable.arch = "x86_64"

            # Prv: module($n)
            solvable.add_deparray(solv.SOLVABLE_PROVIDES,
                                  pool.Dep("module(%s)" % n))
            # Prv: module($n:$s) = $v
            solvable.add_deparray(solv.SOLVABLE_PROVIDES,
                                  pool.Dep("module(%s:%s)" % (n, s)).Rel(
                                      solv.REL_EQ, pool.Dep(str(v))))

            requires = self._deps2reqs(normdeps(mmd, "get_requires"))
            solvable.add_deparray(solv.SOLVABLE_REQUIRES, requires)

            # Con: module($n)
            solvable.add_deparray(solv.SOLVABLE_CONFLICTS, pool.Dep("module(%s)" % n))

            solvables.append(solvable)
        else:
            # Input module
            # Context means two things:
            # * Unique identifier
            # * Offset for the dependency which was used
            normalized_deps = normdeps(mmd, "get_buildrequires")
            for c, deps in enumerate(mmd.get_dependencies()):
                # $n:$s:$c-$v.src
                solvable = self.build_repo.add_solvable()
                solvable.name = "%s:%s:%d:%d" % (n, s, v, c)
                solvable.evr = str(v)
                solvable.arch = "src"

                requires = self._deps2reqs([normalized_deps[c]])
                solvable.add_deparray(solv.SOLVABLE_REQUIRES, requires)

                solvables.append(solvable)

        return solvables

    def solve(self, mmd, policy=MMDResolverPolicy.First):
        """
        Solves dependencies of module defined by `mmd` object. Returns set
        containing frozensets with all the possible combinations which
        satisfied dependencies.

        :return: set of frozensets of n:s:v:c of modules which satisfied the
            dependency solving.
        """
        solvables = self.add_modules(mmd)
        if not solvables:
            raise ValueError("No module(s) found for resolving")
        self.pool.createwhatprovides()

        s2nsvc = lambda s: "%s:%s" % (s.name, s.arch)
        s2ns = lambda s: ":".join(s.name.split(":", 2)[:2])

        solver = self.pool.Solver()
        alternatives = collections.OrderedDict()
        for src in solvables:
            job = self.pool.Job(solv.Job.SOLVER_INSTALL | solv.Job.SOLVER_SOLVABLE, src.id)
            requires = src.lookup_deparray(solv.SOLVABLE_REQUIRES)
            if len(requires) > 1:
                raise SystemError("At max one element should be in Requires: %s" % requires)
            elif len(requires) == 0:
                return set([frozenset([s2nsvc(src)])])

            requires = requires[0]
            src_alternatives = alternatives[src] = collections.OrderedDict()

            # TODO: replace this ugliest workaround ever with sane code of parsing rich deps.
            # We need to split them because whatprovides() treats "and" same way as "or" which is
            # not enough to generate combinations.
            # Source solvables have Req: (X and Y and Z)
            # Binary solvables have Req: ((X and Y) or (X and Z))
            # They do use "or" within "and", so simple string split won't work for binary packages.
            if src.arch != "src":
                raise NotImplementedError
            deps = str(requires).split(" and ")
            if len(deps) > 1:
                deps[0] = deps[0][1:]
                deps[-1] = deps[-1][:-1]
            deps = [self.pool.parserpmrichdep(dep) if dep.startswith("(") else self.pool.Dep(dep)
                    for dep in deps]

            for opt in itertools.product(*[self.pool.whatprovides(dep) for dep in deps]):
                log.debug("Testing %s with combination: %s", src, opt)
                if policy == MMDResolverPolicy.All:
                    kfunc = s2nsvc
                elif policy == MMDResolverPolicy.First:
                    kfunc = s2ns
                key = tuple(kfunc(s) for s in opt)
                alternative = src_alternatives.setdefault(key, [])
                jobs = [self.pool.Job(solv.Job.SOLVER_FAVOR | solv.Job.SOLVER_SOLVABLE, s.id)
                        for s in opt] + [job]
                log.debug("Jobs:")
                for j in jobs:
                    log.debug("  - %s", j)
                problems = solver.solve(jobs)
                if problems:
                    raise RuntimeError("Problems were found during solve(): %s" % ", ".join(
                                       str(p) for p in problems))
                newsolvables = solver.transaction().newsolvables()
                log.debug("Transaction:")
                for s in newsolvables:
                    log.debug("  - %s", s)
                alternative.append(newsolvables)

        if policy == MMDResolverPolicy.First:
            # Prune
            for transactions in alternatives.values():
                for ns, trans in transactions.items():
                    try:
                        transactions[ns] = [next(t for t in trans
                                                 if set(ns) <= set(s2ns(s) for s in t))]
                    except StopIteration:
                        # No transactions found for requested N:S
                        del transactions[ns]
                        continue

        return set(frozenset(s2nsvc(s) for s in transactions[0])
                   for src_alternatives in alternatives.values()
                   for transactions in src_alternatives.values())



# -*- coding: utf-8 -*-
# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com>

from module_build_service import log
from module_build_service.resolver.base import GenericResolver
from module_build_service import models
from module_build_service.errors import UnprocessableEntity


class DBResolver(GenericResolver):
    """
    Resolver using the MBS database
    """
    backend = 'db'

    def __init__(self, config):
        self.config = config

    def get_module_modulemds(self, name, stream, version=None, context=None, strict=False):
        """
        Gets the module modulemds from the resolver.
        :param name: a string of the module's name
        :param stream: a string of the module's stream
        :param version: a string or int of the module's version. When None, latest version will
            be returned.
        :param context: a string of the module's context. When None, all contexts will
            be returned.
        :kwarg strict: Normally this function returns [] if no module can be
            found.  If strict=True, then a UnprocessableEntity is raised.
        :return: List of Modulemd metadata instances matching the query
        """
        with models.make_session(self.config) as session:
            if version and context:
                builds = [models.ModuleBuild.get_build_from_nsvc(
                    session, name, stream, version, context)]
            elif not version and not context:
                builds = models.ModuleBuild.get_last_builds_in_stream(
                    session, name, stream)
            else:
                raise NotImplemented(
                    "This combination of name/stream/version/context is not implemented")

            if not builds and strict:
                raise UnprocessableEntity(
                    "Cannot find any module builds for %s:%s" % (name, stream))
            return [build.mmd() for build in builds]

    def resolve_profiles(self, mmd, keys):
        """
        Returns a dictionary with keys set according the `keys` parameters and values
        set to the union of all components defined in all installation profiles matching
        the key in all buildrequires. If there are some modules loaded by
        utils.load_local_builds(...), these local modules will be considered when returning
        the profiles.
        :param mmd: Modulemd.Module instance representing the module
        :param keys: list of modulemd installation profiles to include in the result
        :return: a dictionary
        """
        results = {}
        for key in keys:
            results[key] = set()
        with models.make_session(self.config) as session:
            for module_name, module_info in mmd.get_xmd()['mbs']['buildrequires'].items():
                local_modules = models.ModuleBuild.local_modules(
                    session, module_name, module_info['stream'])
                if local_modules:
                    local_module = local_modules[0]
                    log.info('Using local module {0!r} to resolve profiles.'.format(local_module))
                    dep_mmd = local_module.mmd()
                    for key in keys:
                        if key in dep_mmd.get_profiles().keys():
                            results[key] |= set(dep_mmd.get_profiles()[key].get_rpms().get())
                    continue

                build = models.ModuleBuild.get_build_from_nsvc(
                    session, module_name, module_info['stream'], module_info['version'],
                    module_info['context'], state=models.BUILD_STATES['ready'])
                if not build:
                    raise UnprocessableEntity('The module {}:{}:{}:{} was not found'.format(
                        module_name, module_info['stream'], module_info['version'],
                        module_info['context']))
                dep_mmd = build.mmd()

                # Take note of what rpms are in this dep's profile
                for key in keys:
                    if key in dep_mmd.get_profiles().keys():
                        results[key] |= set(dep_mmd.get_profiles()[key].get_rpms().get())

        # Return the union of all rpms in all profiles of the given keys
        return results

    def get_module_build_dependencies(self, name=None, stream=None, version=None, context=None,
                                      mmd=None, strict=False):
        """
        Returns a dictionary of koji_tag:mmd of all the dependencies
        :kwarg name: a string of a module's name (required if mmd is not set)
        :kwarg stream: a string of a module's stream (required if mmd is not set)
        :kwarg version: a string of a module's version (required if mmd is not set)
        :kwarg context: a string of a module's context (required if mmd is not set)
        :kwarg mmd: Modulemd.Module object. If this is set, the mmd will be used instead of
            querying the DB with the name, stream, version, and context.
        :kwarg strict: Normally this function returns None if no module can be
            found.  If strict=True, then an UnprocessableEntity is raised.
        :return: a dictionary
        """
        if mmd:
            log.debug('get_module_build_dependencies(mmd={0!r} strict={1!r})'.format(mmd, strict))
        elif any(x is None for x in [name, stream, version, context]):
            raise RuntimeError('The name, stream, version, and/or context weren\'t specified')
        else:
            version = str(version)
            log.debug('get_module_build_dependencies({0}, strict={1!r})'.format(
                ', '.join([name, stream, str(version), context]), strict))

        module_tags = {}
        with models.make_session(self.config) as session:
            if mmd:
                queried_mmd = mmd
                nsvc = ':'.join([
                    mmd.get_name(), mmd.get_stream(), str(mmd.get_version()),
                    mmd.get_context() or '00000000'])
            else:
                build = models.ModuleBuild.get_build_from_nsvc(
                    session, name, stream, version, context)
                if not build:
                    raise UnprocessableEntity('The module {} was not found'.format(
                        ':'.join([name, stream, version, context])))
                queried_mmd = build.mmd()
                nsvc = ':'.join([name, stream, version, context])

            xmd_mbs = queried_mmd.get_xmd().get('mbs')
            if not xmd_mbs or 'buildrequires' not in xmd_mbs.keys():
                raise RuntimeError(
                    'The module {} did not contain its modulemd or did not have '
                    'its xmd attribute filled out in MBS'.format(nsvc))

            buildrequires = xmd_mbs['buildrequires']
            for br_name, details in buildrequires.items():
                build = models.ModuleBuild.get_build_from_nsvc(
                    session, br_name, details['stream'], details['version'], details['context'],
                    state=models.BUILD_STATES['ready'])
                if not build:
                    raise RuntimeError(
                        'Buildrequired module %s %r does not exist in MBS db' % (br_name, details))
                module_tags[build.koji_tag] = build.mmd()

        return module_tags

    def resolve_requires(self, requires):
        """
        Resolves the requires list of N:S or N:S:V:C to a dictionary with keys as
        the module name and the values as a dictionary with keys of ref,
        stream, version, filtered_rpms.
        If there are some modules loaded by utils.load_local_builds(...), these
        local modules will be considered when resolving the requires. A RuntimeError
        is raised on DB lookup errors.
        :param requires: a dictionary with the module name as the key and the stream as the value
        :return: a dictionary
        """
        new_requires = {}
        with models.make_session(self.config) as session:
            for nsvc in requires:
                nsvc_splitted = nsvc.split(":")
                if len(nsvc_splitted) == 2:
                    module_name, module_stream = nsvc_splitted
                    module_version = None
                    module_context = None
                elif len(nsvc_splitted) == 4:
                    module_name, module_stream, module_version, module_context = nsvc_splitted
                else:
                    raise ValueError(
                        "Only N:S or N:S:V:C is accepted by resolve_requires, got %s" % nsvc)

                local_modules = models.ModuleBuild.local_modules(
                    session, module_name, module_stream)
                if local_modules:
                    local_build = local_modules[0]
                    new_requires[module_name] = {
                        'ref': None,
                        'stream': local_build.stream,
                        'version': local_build.version,
                        'context': local_build.context,
                        # No need to set filtered_rpms for local builds, because MBS
                        # filters the RPMs automatically when the module build is
                        # done.
                        'filtered_rpms': []
                    }
                    continue

                if module_version is None or module_context is None:
                    build = models.ModuleBuild.get_last_build_in_stream(
                        session, module_name, module_stream)
                else:
                    build = models.ModuleBuild.get_build_from_nsvc(
                        session, module_name, module_stream, module_version, module_context)

                if not build:
                    raise UnprocessableEntity('The module {} was not found'.format(nsvc))

                commit_hash = None
                filtered_rpms = []
                mmd = build.mmd()
                mbs_xmd = mmd.get_xmd().get('mbs')
                if mbs_xmd and 'commit' in mbs_xmd.keys():
                    commit_hash = mbs_xmd['commit']
                else:
                    raise RuntimeError(
                        'The module "{0}" didn\'t contain a commit hash in its xmd'
                        .format(module_name))

                if "mse" not in mbs_xmd.keys() or not mbs_xmd["mse"]:
                    raise RuntimeError(
                        'The module "{}" is not built using Module Stream Expansion. '
                        'Please rebuild this module first'.format(nsvc))

                # Find out the particular NVR of filtered packages
                rpm_filter = mmd.get_rpm_filter()
                if rpm_filter and rpm_filter.get():
                    for rpm in build.component_builds:
                        if rpm.package in rpm_filter.get():
                            filtered_rpms.append(rpm.nvr)

                new_requires[module_name] = {
                    'ref': commit_hash,
                    'stream': module_stream,
                    'version': build.version,
                    'context': build.context,
                    'filtered_rpms': filtered_rpms,
                }

        return new_requires



"""Add component_builds.weight

Revision ID: f5b1c5203cce
Revises: 524c3b1c683c
Create Date: 2017-11-08 17:18:27.401546

"""

# revision identifiers, used by Alembic.
revision = 'f5b1c5203cce'
down_revision = '524c3b1c683c'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.add_column('component_builds', sa.Column('weight', sa.Float(), nullable=True))


def downgrade():
    op.drop_column('component_builds', 'weight')



# -*- coding: utf-8 -*-


# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Petr Šabata <contyk@redhat.com>
#            Filip Valder <fvalder@redhat.com>

import imp
import os
import pkg_resources
import re
import sys

from module_build_service import logger


# TODO: It'd be nice to reuse this from models.ModuleBuild.rebuild_strategies but models.py
# currently relies on this file, so we can't import it
SUPPORTED_STRATEGIES = ['changed-and-after', 'only-changed', 'all']

SUPPORTED_RESOLVERS = {
    'mbs': {'builders': ['mock']},
    'db': {'builders': ['koji', 'mock', 'copr']}
}


def init_config(app):
    """ Configure MBS and the Flask app
    """
    config_module = None
    config_file = '/etc/module-build-service/config.py'
    config_section = 'DevConfiguration'

    # automagically detect production environment:
    #   - existing and readable config_file presets ProdConfiguration
    try:
        with open(config_file):
            config_section = 'ProdConfiguration'
    except Exception:
        pass
    #   - Flask app within mod_wsgi presets ProdConfiguration
    flask_app_env = hasattr(app, 'request') and hasattr(app.request, 'environ')
    if flask_app_env and any([var.startswith('mod_wsgi.')
                              for var in app.request.environ]):
        config_section = 'ProdConfiguration'

    # Load LocalBuildConfiguration section in case we are building modules
    # locally.
    if "build_module_locally" in sys.argv:
        config_section = "LocalBuildConfiguration"

    # try getting config_file from os.environ
    if 'MBS_CONFIG_FILE' in os.environ:
        config_file = os.environ['MBS_CONFIG_FILE']
    # try getting config_section from os.environ
    if 'MBS_CONFIG_SECTION' in os.environ:
        config_section = os.environ['MBS_CONFIG_SECTION']
    # preferably get these values from Flask app
    if flask_app_env:
        # try getting config_file from Flask app
        if 'MBS_CONFIG_FILE' in app.request.environ:
            config_file = app.request.environ['MBS_CONFIG_FILE']
        # try getting config_section from Flask app
        if 'MBS_CONFIG_SECTION' in app.request.environ:
            config_section = app.request.environ['MBS_CONFIG_SECTION']
    # TestConfiguration shall only be used for running tests, otherwise...
    if any(['py.test' in arg or 'pytest' in arg for arg in sys.argv]):
        config_section = 'TestConfiguration'
        from conf import config
        config_module = config
    # ...MODULE_BUILD_SERVICE_DEVELOPER_ENV has always the last word
    # and overrides anything previously set before!
    # Again, check Flask app (preferably) or fallback to os.environ.
    # In any of the following cases, use configuration directly from MBS package
    # -> /conf/config.py.
    elif (flask_app_env and
          'MODULE_BUILD_SERVICE_DEVELOPER_ENV' in app.request.environ):
        if app.request.environ['MODULE_BUILD_SERVICE_DEVELOPER_ENV'].lower() in (
                '1', 'on', 'true', 'y', 'yes'):
            config_section = 'DevConfiguration'
            from conf import config
            config_module = config
    elif ('MODULE_BUILD_SERVICE_DEVELOPER_ENV' in os.environ and
          os.environ['MODULE_BUILD_SERVICE_DEVELOPER_ENV'].lower() in (
            '1', 'on', 'true', 'y', 'yes')):
        config_section = 'DevConfiguration'
        from conf import config
        config_module = config
    # try loading configuration from file
    if not config_module:
        try:
            config_module = imp.load_source('mbs_runtime_config',
                                            config_file)
        except Exception:
            raise SystemError("Configuration file {} was not found."
                              .format(config_file))

    # finally configure MBS and the Flask app
    config_section_obj = getattr(config_module, config_section)
    conf = Config(config_section_obj)
    app.config.from_object(config_section_obj)
    return conf


class Path:
    """
    Config type for paths. Expands the users home directory.
    """
    pass


class Config(object):
    """Class representing the orchestrator configuration."""
    _defaults = {
        'debug': {
            'type': bool,
            'default': False,
            'desc': 'Debug mode'},
        'system': {
            'type': str,
            'default': 'koji',
            'desc': 'The buildsystem to use.'},
        'db': {
            'type': str,
            'default': '',
            'desc': 'RDB URL.'},
        'default_dist_tag_prefix': {
            'type': str,
            'default': 'module+',
            'desc': 'Default dist-tag prefix for built modules.'},
        'polling_interval': {
            'type': int,
            'default': 0,
            'desc': 'Polling interval, in seconds.'},
        'cache_dir': {
            'type': Path,
            'default': '~/modulebuild/cache',
            'desc': 'Cache directory'},
        'mbs_url': {
            'type': str,
            'default': 'https://mbs.fedoraproject.org/module-build-service/1/module-builds/',
            'desc': 'MBS instance url for MBSResolver'},
        'koji_config': {
            'type': str,
            'default': None,
            'desc': 'Koji config file.'},
        'koji_profile': {
            'type': str,
            'default': None,
            'desc': 'Koji config profile.'},
        'koji_arches': {
            'type': list,
            'default': [],
            'desc': 'Koji architectures.'},
        'koji_build_priority': {
            'type': int,
            'default': 10,
            'desc': ''},
        'koji_repository_url': {
            'type': str,
            'default': None,
            'desc': 'Koji repository URL.'},
        'koji_build_macros_target': {
            'type': str,
            'default': '',
            'desc': 'Target to build "module-build-macros" RPM in.'},
        'koji_tag_prefixes': {
            'type': list,
            'default': ['module'],
            'desc': 'List of allowed koji tag prefixes.'},
        'koji_target_delete_time': {
            'type': int,
            'default': 24 * 3600,
            'desc': 'Time in seconds after which the Koji target of '
                    'built module is deleted'},
        'koji_enable_content_generator': {
            'type': bool,
            'default': True,
            'desc': 'Enable or disable imports to koji using content '
                    'generator api'},
        'allow_custom_scmurls': {
            'type': bool,
            'default': False,
            'desc': 'Allow custom scmurls.'},
        'rpms_default_repository': {
            'type': str,
            'default': 'git://pkgs.fedoraproject.org/rpms/',
            'desc': 'RPMs default repository URL.'},
        'rpms_allow_repository': {
            'type': bool,
            'default': False,
            'desc': 'Allow custom RPMs repositories.'},
        'rpms_default_cache': {
            'type': str,
            'default': 'http://pkgs.fedoraproject.org/repo/pkgs/',
            'desc': 'RPMs default cache URL.'},
        'rpms_allow_cache': {
            'type': bool,
            'default': False,
            'desc': 'Allow custom RPMs cache.'},
        'modules_default_repository': {
            'type': str,
            'default': 'git://pkgs.fedoraproject.org/modules/',
            'desc': 'Included modules default repository URL.'},
        'modules_allow_repository': {
            'type': bool,
            'default': False,
            'desc': 'Allow custom included modules repositories.'},
        'allowed_groups': {
            'type': set,
            'default': set(['packager']),
            'desc': 'The set of groups allowed to submit builds.'},
        'log_backend': {
            'type': str,
            'default': None,
            'desc': 'Log backend'},
        'log_file': {
            'type': str,
            'default': '',
            'desc': 'Path to log file'},
        'log_level': {
            'type': str,
            'default': 0,
            'desc': 'Log level'},
        'build_logs_dir': {
            'type': Path,
            'default': "",
            'desc': 'Directory to store module build logs to.'},
        'build_logs_name_format': {
            'type': str,
            'default': "build-{id}.log",
            'desc': ('Format of a module build log\'s name. Use `Build` attributes as formatting '
                     'kwargs')},
        'krb_keytab': {
            'type': None,
            'default': None,
            'desc': ''},
        'krb_principal': {
            'type': None,
            'default': None,
            'desc': ''},
        'krb_ccache': {
            'type': None,
            'default': '/tmp/krb5cc_module_build_service',
            'desc': ''},
        'messaging': {
            'type': str,
            'default': 'fedmsg',
            'desc': 'The messaging system to use.'},
        'messaging_topic_prefix': {
            'type': list,
            'default': ['org.fedoraproject.prod'],
            'desc': 'The messaging system topic prefixes which we are interested in.'},
        'amq_recv_addresses': {
            'type': list,
            'default': [],
            'desc': 'Apache MQ broker url to receive messages.'},
        'amq_dest_address': {
            'type': str,
            'default': '',
            'desc': 'Apache MQ broker address to send messages'},
        'amq_cert_file': {
            'type': str,
            'default': '',
            'desc': 'Certificate for Apache MQ broker auth.'},
        'amq_private_key_file': {
            'type': str,
            'default': '',
            'desc': 'Private key for Apache MQ broker auth.'},
        'amq_trusted_cert_file': {
            'type': str,
            'default': '',
            'desc': 'Trusted certificate for ssl connection.'},
        'distgits': {
            'type': dict,
            'default': {
                'git://pkgs.fedoraproject.org': ('fedpkg clone --anonymous {}',
                                                 'fedpkg --release module sources'),
            },
            'desc': 'Mapping between dist-git and command to '},
        'mock_config': {
            'type': str,
            'default': 'fedora-25-x86_64',
            'desc': ''},
        'mock_config_file': {
            'type': list,
            'default': ['/etc/module-build-service/mock.cfg', 'conf/mock.cfg'],
            'desc': 'List of mock config file paths in order of preference.'},
        'mock_build_srpm_cmd': {
            'type': str,
            'default': 'fedpkg --release f26 srpm',
            'desc': ''},
        'mock_resultsdir': {
            'type': Path,
            'default': '~/modulebuild/builds',
            'desc': 'Directory for Mock build results.'},
        'mock_purge_useless_logs': {
            'type': bool,
            'default': True,
            'desc': 'Remove empty or otherwise useless log files.'},
        'arch_autodetect': {
            'type': bool,
            'default': True,
            'desc': 'Auto-detect machine arch when configuring builder.'},
        'arch_fallback': {
            'type': str,
            'default': 'x86_64',
            'desc': 'Fallback arch if auto-detection is off or unable to determine it.'},
        'scmurls': {
            'type': list,
            'default': [],
            'desc': 'Allowed SCM URLs for submitted module.'},
        'yaml_submit_allowed': {
            'type': bool,
            'default': False,
            'desc': 'Is it allowed to directly submit modulemd yaml file?'},
        'num_concurrent_builds': {
            'type': int,
            'default': 0,
            'desc': 'Number of concurrent component builds.'},
        'net_timeout': {
            'type': int,
            'default': 120,
            'desc': 'Global network timeout for read/write operations, in seconds.'},
        'net_retry_interval': {
            'type': int,
            'default': 30,
            'desc': 'Global network retry interval for read/write operations, in seconds.'},
        'scm_net_timeout': {
            'type': int,
            'default': 60,
            'desc': 'Network timeout for SCM operations, in seconds.'},
        'scm_net_retry_interval': {
            'type': int,
            'default': 15,
            'desc': 'Network retry interval for SCM operations, in seconds.'},
        'no_auth': {
            'type': bool,
            'default': False,
            'desc': 'Disable client authentication.'},
        'admin_groups': {
            'type': set,
            'default': set([]),
            'desc': 'The set of groups allowed to manage MBS.'},
        'yum_config_file': {
            'type': list,
            'default': ['/etc/module-build-service/yum.conf', 'conf/yum.conf'],
            'desc': 'List of yum config file paths in order of preference.'},
        'auth_method': {
            'type': str,
            'default': 'oidc',
            'desc': 'Authentiation method to MBS. Options are oidc or kerberos'},
        'kerberos_http_host': {
            'type': str,
            'default': '',
            'desc': ('Hardcodes the HTTP host MBS identifies as in Kerberos. If this isn\'t set, '
                     'it will be derived dynamically.')},
        'kerberos_keytab': {
            'type': str,
            'default': '',
            'desc': ('Overrides the use of the environment variable KRB5_KTNAME, which specifies '
                     'the location to the Kerberos keytab for authentication.')},
        'ldap_uri': {
            'type': str,
            'default': '',
            'desc': 'LDAP URI to query for group information when using Kerberos authentication'},
        'ldap_groups_dn': {
            'type': str,
            'default': '',
            'desc': ('The distinguished name of the container or organizational unit containing '
                     'the groups in LDAP')},
        'base_module_names': {
            'type': set,
            'default': set(['platform', 'bootstrap']),
            'desc': ("Set of module names which defines the product version "
                     "(by their stream) of modules depending on them.")},
        'koji_cg_build_tag_template': {
            'type': str,
            'default': "{}-modular-updates-candidate",
            'desc': "Name of a Koji tag where the top-level Content Generator "
                    "build is tagged to. The '{}' string is replaced by a "
                    "stream name of a base module on top of which the "
                    "module is built."},
        'koji_cg_default_build_tag': {
            'type': str,
            'default': "modular-updates-candidate",
            'desc': "The name of Koji tag which should be used as fallback "
                    "when koji_cg_build_tag_template tag is not found in "
                    "Koji."},
        'rebuild_strategy': {
            'type': str,
            'default': 'changed-and-after',
            'desc': 'The module rebuild strategy to use by default.'},
        'rebuild_strategy_allow_override': {
            'type': bool,
            'default': False,
            'desc': ('Allows a user to specify the rebuild strategy they want to use when '
                     'submitting a module build.')},
        'rebuild_strategies_allowed': {
            'type': list,
            'default': SUPPORTED_STRATEGIES,
            'desc': ('The allowed module rebuild strategies. This is only used when '
                     'REBUILD_STRATEGY_ALLOW_OVERRIDE is True.')},
        'cleanup_failed_builds_time': {
            'type': int,
            'default': 180,
            'desc': ('Time in days when to cleanup failed module builds and transition them to '
                     'the "garbage" state.')},
        'resolver': {
            'type': str,
            'default': 'db',
            'desc': 'Where to look up for modules. Note that this can (and '
                    'probably will) be builder-specific.'},
    }

    def __init__(self, conf_section_obj):
        """
        Initialize the Config object with defaults and then override them
        with runtime values.
        """

        # set defaults
        for name, values in self._defaults.items():
            self.set_item(name, values['default'], values['type'])

        # override defaults
        for key in dir(conf_section_obj):
            # skip keys starting with underscore
            if key.startswith('_'):
                continue
            # set item (lower key)
            self.set_item(key.lower(), getattr(conf_section_obj, key))

    def set_item(self, key, value, value_type=None):
        """
        Set value for configuration item. Creates the self._key = value
        attribute and self.key property to set/get/del the attribute.
        """
        if key == 'set_item' or key.startswith('_'):
            raise Exception("Configuration item's name is not allowed: %s" % key)

        # Create the empty self._key attribute, so we can assign to it.
        if not hasattr(self, "_" + key):
            setattr(self, "_" + key, None)

            # Create self.key property to access the self._key attribute.
            # Use the setifok_func if available for the attribute.
            setifok_func = '_setifok_{}'.format(key)
            if hasattr(self, setifok_func):
                setx = lambda self, val: getattr(self, setifok_func)(val)
            elif value_type == Path:
                # For paths, expanduser.
                setx = lambda self, val: setattr(
                    self, "_" + key, os.path.expanduser(val))
            else:
                setx = lambda self, val: setattr(self, "_" + key, val)
            getx = lambda self: getattr(self, "_" + key)
            delx = lambda self: delattr(self, "_" + key)
            setattr(Config, key, property(getx, setx, delx))

        # managed/registered configuration items
        if key in self._defaults:
            # type conversion for configuration item
            convert = self._defaults[key]['type']
            if convert in [bool, int, list, str, set, dict]:
                try:
                    # Do no try to convert None...
                    if value is not None:
                        value = convert(value)
                except Exception:
                    raise TypeError("Configuration value conversion failed for name: %s" % key)
            # unknown type/unsupported conversion, or conversion not needed
            elif convert is not None and convert not in [Path]:
                raise TypeError("Unsupported type %s for configuration item name: %s"
                                % (convert, key))

        # Set the attribute to the correct value
        setattr(self, key, value)

    #
    # Register your _setifok_* handlers here
    #

    def _setifok_system(self, s):
        s = str(s)
        if s not in ("koji", "mock"):
            raise ValueError("Unsupported buildsystem: %s." % s)
        self._system = s

    def _setifok_polling_interval(self, i):
        if not isinstance(i, int):
            raise TypeError("polling_interval needs to be an int")
        if i < 0:
            raise ValueError("polling_interval must be >= 0")
        self._polling_interval = i

    def _setifok_rpms_default_repository(self, s):
        rpm_repo = str(s)
        rpm_repo = rpm_repo.rstrip('/') + '/'
        self._rpms_default_repository = rpm_repo

    def _setifok_rpms_default_cache(self, s):
        rpm_cache = str(s)
        rpm_cache = rpm_cache.rstrip('/') + '/'
        self._rpms_default_cache = rpm_cache

    def _setifok_log_backend(self, s):
        if s is None:
            self._log_backend = "console"
        elif s not in logger.supported_log_backends():
            raise ValueError("Unsupported log backend")
        self._log_backend = str(s)

    def _setifok_log_file(self, s):
        if s is None:
            self._log_file = ""
        else:
            self._log_file = str(s)

    def _setifok_log_level(self, s):
        level = str(s).lower()
        self._log_level = logger.str_to_log_level(level)

    def _setifok_messaging(self, s):
        """ Validate that the specified messaging backend corresponds with one
        of the installed plugins.  The MBS core provides two such plugins, but
        a third-party could install another usable one.
        """
        entrypoints = pkg_resources.iter_entry_points('mbs.messaging_backends')
        installed_backends = [e.name for e in entrypoints]
        s = str(s)
        if s not in installed_backends:
            raise ValueError('The messaging plugin for "{0}" is not installed.'
                             ' The following are installed: {1}'
                             .format(s, ', '.join(installed_backends)))
        self._messaging = s

    def _setifok_amq_recv_addresses(self, l):
        assert isinstance(l, list) or isinstance(l, tuple)
        self._amq_recv_addresses = list(l)

    def _setifok_scmurls(self, l):
        if not isinstance(l, list):
            raise TypeError("scmurls needs to be a list.")
        self._scmurls = [str(x) for x in l]

    def _setifok_num_concurrent_builds(self, i):
        if not isinstance(i, int):
            raise TypeError('NUM_CONCURRENT_BUILDS needs to be an int')
        if i < 0:
            raise ValueError('NUM_CONCURRENT_BUILDS must be >= 0')
        self._num_concurrent_builds = i

    def _setifok_auth_method(self, s):
        s = str(s)
        if s.lower() not in ('oidc', 'kerberos'):
            raise ValueError('Unsupported authentication method')
        if s.lower() == 'kerberos':
            try:
                import ldap3  # noqa
            except ImportError:
                raise ValueError("ldap3 is required for kerberos authz")
        self._auth_method = s.lower()

    def _setifok_kerberos_keytab(self, s):
        keytab = str(s)
        if keytab:
            keytab = os.path.expanduser(keytab)
            if not os.path.exists(keytab):
                raise ValueError('The path set for KERBEROS_KEYTAB does not exist')

        self._kerberos_keytab = keytab

    def _setifok_ldap_uri(self, s):
        ldap_uri = str(s)

        if ldap_uri and not re.match(r'^(?:ldap(?:s)?:\/\/.+)$', ldap_uri):
            raise ValueError('LDAP_URI is invalid. It must start with "ldap://" or "ldaps://"')

        self._ldap_uri = ldap_uri

    def _setifok_rebuild_strategy(self, strategy):
        if strategy not in SUPPORTED_STRATEGIES:
            raise ValueError('The strategy "{0}" is not supported. Choose from: {1}'
                             .format(strategy, ', '.join(SUPPORTED_STRATEGIES)))
        self._rebuild_strategy = strategy

    def _setifok_rebuild_strategies_allowed(self, strategies):
        if not isinstance(strategies, list):
            raise ValueError('REBUILD_STRATEGIES_ALLOWED must be a list')
        elif not strategies:
            raise ValueError('REBUILD_STRATEGIES_ALLOWED must contain at least one rebuild '
                             'strategy')
        for strategy in strategies:
            if strategy not in SUPPORTED_STRATEGIES:
                raise ValueError('REBUILD_STRATEGIES_ALLOWED must be one of: {0}'
                                 .format(', '.join(SUPPORTED_STRATEGIES)))

        self._rebuild_strategies_allowed = strategies

    def _setifok_cleanup_failed_builds_time(self, num_days):
        if num_days < 1:
            raise ValueError('CLEANUP_FAILED_BUILDS_TIME must be set to 1 or more days')
        self._cleanup_failed_builds_time = num_days

    def _setifok_resolver(self, s):
        if s not in SUPPORTED_RESOLVERS.keys():
            raise ValueError('The resolver "{0}" is not supported. Choose from: {1}'
                             .format(s, ', '.join(SUPPORTED_RESOLVERS.keys())))
        self._resolver = s



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Ralph Bean <rbean@redhat.com>

""" Handlers for repo change events on the message bus. """

import module_build_service.builder
import logging
import koji
from module_build_service import models, log
from module_build_service.utils import start_next_batch_build

logging.basicConfig(level=logging.DEBUG)


def done(config, session, msg):
    """ Called whenever koji rebuilds a repo, any repo. """

    # First, find our ModuleBuild associated with this repo, if any.
    tag = msg.repo_tag
    if config.system in ('koji', 'test') and not tag.endswith('-build'):
        log.debug("Tag %r does not end with '-build' suffix, ignoring" % tag)
        return
    tag = tag[:-6] if tag.endswith('-build') else tag
    module_build = models.ModuleBuild.from_repo_done_event(session, msg)
    if not module_build:
        log.debug("No module build found associated with koji tag %r" % tag)
        return

    # It is possible that we have already failed.. but our repo is just being
    # routinely regenerated.  Just ignore that.  If module_build_service says the module is
    # dead, then the module is dead.
    if module_build.state == models.BUILD_STATES['failed']:
        log.info("Ignoring repo regen for already failed %r" % module_build)
        return

    # Get the list of untagged components in current/previous batches which
    # have been built successfully
    if config.system in ('koji', 'test') and module_build.component_builds:
        untagged_components = [
            c for c in module_build.up_to_current_batch()
            if (not c.tagged or (not c.tagged_in_final and not c.build_time_only)) and
            c.state == koji.BUILD_STATES['COMPLETE']
        ]
        if untagged_components:
            log.info("Ignoring repo regen, because not all components are tagged.")
            return
        if all([c.state is None for c in module_build.current_batch()]):
            log.info("Ignoring repo regen because no components have started in the batch.")
            return

    # If there are no components in this module build, then current_batch will be empty
    if not module_build.component_builds:
        current_batch = []
    else:
        current_batch = module_build.current_batch()

    # If any in the current batch are still running.. just wait.
    running = [c.state == koji.BUILD_STATES['BUILDING'] for c in current_batch]
    if any(running):
        log.info(
            "%r has %r of %r components still "
            "building in this batch (%r total)" % (
                module_build, len(running), len(current_batch),
                len(module_build.component_builds)))
        return

    # Assemble the list of all successful components in the batch.
    good = [c for c in current_batch if c.state == koji.BUILD_STATES['COMPLETE']]

    # If *none* of the components completed for this batch, then obviously the
    # module fails.  However!  We shouldn't reach this scenario.  There is
    # logic over in the component handler which should fail the module build
    # first before we ever get here.  This is here as a race condition safety
    # valve.
    if module_build.component_builds and not good:
        module_build.transition(config, models.BUILD_STATES['failed'],
                                "Some components failed to build.")
        session.commit()
        log.warn("Odd!  All components in batch failed for %r." % module_build)
        return

    groups = module_build_service.builder.GenericBuilder.default_buildroot_groups(
        session, module_build)

    builder = module_build_service.builder.GenericBuilder.create(
        module_build.owner, module_build, config.system, config,
        tag_name=tag, components=[c.package for c in module_build.component_builds])
    builder.buildroot_connect(groups)

    # If we have reached here then we know the following things:
    #
    # - All components in this batch have finished (failed or succeeded)
    # - One or more succeeded.
    # - They have been regenerated back into the buildroot.
    #
    # So now we can either start a new batch if there are still some to build
    # or, if everything is built successfully, then we can bless the module as
    # complete.
    has_unbuilt_components = False
    has_failed_components = False
    for c in module_build.component_builds:
        if c.state in [None, koji.BUILD_STATES['BUILDING']]:
            has_unbuilt_components = True
        elif (c.state in [koji.BUILD_STATES['FAILED'],
                          koji.BUILD_STATES['CANCELED']]):
            has_failed_components = True

    further_work = []
    if has_unbuilt_components and not has_failed_components:
        # Ok, for the subset of builds that did complete successfully, check to
        # see if they are in the buildroot before starting new batch.
        artifacts = [component_build.nvr for component_build in good]
        if not builder.buildroot_ready(artifacts):
            log.info("Not all of %r are in the buildroot.  Waiting." % artifacts)
            return

        # Try to start next batch build, because there are still unbuilt
        # components in a module.
        further_work += start_next_batch_build(
            config, module_build, session, builder)

    else:
        if has_failed_components:
            module_build.transition(config, state=models.BUILD_STATES['failed'],
                                    state_reason="Some components failed to build.")
        else:
            module_build.transition(config, state=models.BUILD_STATES['done'])
        session.commit()

    return further_work



# -*- coding: utf-8 -*-
# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Matt Prahl <mprahl@redhat.com> except for the test functions

from __future__ import print_function
from flask_script import Manager
from functools import wraps
import flask_migrate
import logging
import os
import getpass

from werkzeug.datastructures import FileStorage
from module_build_service import app, conf, db, create_app
from module_build_service import models
from module_build_service.utils import (
    submit_module_build_from_yaml,
    load_local_builds, load_mmd, import_mmd
)
from module_build_service.errors import StreamAmbigous
import module_build_service.messaging
import module_build_service.scheduler.consumer


manager = Manager(create_app)
help_args = ('-?', '--help')
manager.help_args = help_args
migrate = flask_migrate.Migrate(app, db)
manager.add_command('db', flask_migrate.MigrateCommand)
manager.add_option('-d', '--debug', dest='debug', action='store_true')
manager.add_option('-v', '--verbose', dest='verbose', action='store_true')
manager.add_option('-q', '--quiet', dest='quiet', action='store_true')


def console_script_help(f):
    @wraps(f)
    def wrapped(*args, **kwargs):
        import sys
        if any([arg in help_args for arg in sys.argv[1:]]):
            command = os.path.basename(sys.argv[0])
            print("""{0}

Usage: {0} [{1}]

See also:
  mbs-manager(1)""".format(command,
                           '|'.join(help_args)))
            sys.exit(2)
        r = f(*args, **kwargs)
        return r
    return wrapped


@console_script_help
@manager.command
def upgradedb():
    """ Upgrades the database schema to the latest revision
    """
    app.config["SERVER_NAME"] = 'localhost'
    # TODO: configurable?
    migrations_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                                  'migrations')
    with app.app_context():
        flask_migrate.upgrade(directory=migrations_dir)


@manager.command
def cleardb():
    """ Clears the database
    """
    models.ModuleBuild.query.delete()
    models.ComponentBuild.query.delete()


@console_script_help
@manager.command
def import_module(mmd_file):
    """ Imports the module from mmd_file
    """
    mmd = load_mmd(mmd_file, is_file=True)
    import_mmd(db.session, mmd)


@manager.option('--stream', action='store', dest="stream")
@manager.option('--file', action='store', dest="yaml_file")
@manager.option('--skiptests', action='store_true', dest="skiptests")
@manager.option('-l', '--add-local-build', action='append', default=None, dest='local_build_nsvs')
@manager.option('-s', '--set-stream', action='append', default=[], dest='default_streams')
def build_module_locally(local_build_nsvs=None, yaml_file=None, stream=None, skiptests=False,
                         default_streams=None):
    """ Performs local module build using Mock
    """
    if 'SERVER_NAME' not in app.config or not app.config['SERVER_NAME']:
        app.config["SERVER_NAME"] = 'localhost'

        if app.config['RESOLVER'] == 'db':
            raise ValueError("Please set RESOLVER to 'mbs' in your "
                             "configuration for local builds.")

    with app.app_context():
        conf.set_item("system", "mock")

        # Use our own local SQLite3 database.
        confdir = os.path.abspath(os.getcwd())
        dbdir = os.path.abspath(os.path.join(confdir, '..')) if confdir.endswith('conf') \
            else confdir
        dbpath = '/{0}'.format(os.path.join(dbdir, '.mbs_local_build.db'))
        dburi = 'sqlite://' + dbpath
        app.config["SQLALCHEMY_DATABASE_URI"] = dburi
        conf.set_item("sqlalchemy_database_uri", dburi)
        if os.path.exists(dbpath):
            os.remove(dbpath)

        db.create_all()
        load_local_builds(local_build_nsvs)

        optional_params = {}
        optional_params["local_build"] = True
        optional_params["default_streams"] = {}
        for ns in default_streams:
            name, stream = ns.split(":")
            optional_params["default_streams"][name] = stream

        username = getpass.getuser()
        if not yaml_file or not yaml_file.endswith(".yaml"):
            raise IOError("Provided modulemd file is not a yaml file.")

        yaml_file_path = os.path.abspath(yaml_file)
        with open(yaml_file_path) as fd:
            filename = os.path.basename(yaml_file)
            handle = FileStorage(fd)
            handle.filename = filename
            try:
                submit_module_build_from_yaml(
                    username, handle, str(stream), skiptests, optional_params)
            except StreamAmbigous as e:
                logging.error(str(e))
                logging.error(
                    "Use '-s module_name:module_stream' to choose the stream")
                return

        stop = module_build_service.scheduler.make_simple_stop_condition(db.session)

        # Run the consumer until stop_condition returns True
        module_build_service.scheduler.main([], stop)


@console_script_help
@manager.command
def run(host=None, port=None, debug=None):
    """ Runs the Flask app, locally.
    """
    host = host or conf.host
    port = port or conf.port
    debug = debug or conf.debug

    logging.info('Starting Module Build Service frontend')

    app.run(
        host=host,
        port=port,
        debug=debug
    )


def manager_wrapper():
    manager.run()


if __name__ == "__main__":
    manager_wrapper()



# -*- coding: utf-8 -*-


# Copyright (c) 2016  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Modified by:
# Written by Karsten Hopp <karsten@redhat.com>
#            Petr Šabata <contyk@redhat.com>

"""SCM handler functions."""

import os
import subprocess as sp
import re
import tempfile
import shutil
import datetime

from module_build_service import log, conf
from module_build_service.errors import (
    Forbidden, ValidationError, UnprocessableEntity, ProgrammingError)
from module_build_service.utils.general import scm_url_schemes, retry


class SCM(object):
    "SCM abstraction class"

    # Assuming git for HTTP schemas
    types = scm_url_schemes()

    def __init__(self, url, branch=None, allowed_scm=None, allow_local=False):
        """Initialize the SCM object using the specified scmurl.

        If url is not in the list of allowed_scm, an error will be raised.

        :param str url: The unmodified scmurl
        :param list allowed_scm: The list of allowed SCMs, optional
        :raises: Forbidden or ValidationError
        """

        if allowed_scm:
            if not (url.startswith(tuple(allowed_scm)) or
                    (allow_local and url.startswith("file://"))):
                raise Forbidden(
                    '%s is not in the list of allowed SCMs' % url)

        # If we are given the option for the git protocol or the http(s) protocol,
        # then just use http(s)
        if re.match(r'(git\+http(?:s)?:\/\/)', url):
            url = url[4:]
        url = url.rstrip('/')

        self.url = url
        self.sourcedir = None

        # once we have more than one SCM provider, we will need some more
        # sophisticated lookup logic
        for scmtype, schemes in SCM.types.items():
            if self.url.startswith(schemes):
                self.scheme = scmtype
                break
        else:
            raise ValidationError('Invalid SCM URL: %s' % url)

        # git is the only one supported SCM provider atm
        if self.scheme == "git":
            match = re.search(r"^(?P<repository>.*/(?P<name>[^?]*))(\?#(?P<commit>.*))?", url)
            self.repository = match.group("repository")
            self.name = match.group("name")
            self.repository_root = self.repository[:-len(self.name)]
            if self.name.endswith(".git"):
                self.name = self.name[:-4]
            self.commit = match.group("commit")
            self.branch = branch if branch else "master"
            if not self.commit:
                self.commit = self.get_latest(self.branch)
            self.version = None
        else:
            raise ValidationError("Unhandled SCM scheme: %s" % self.scheme)

    def verify(self):
        """
        Verifies that the information provided by a user in SCM URL and branch
        matches the information in SCM repository. For example verifies that
        the commit hash really belongs to the provided branch.

        :raises ValidationError
        """
        if not self.sourcedir:
            raise ProgrammingError("Do .checkout() first.")

        found = False
        branches = SCM._run(["git", "branch", "-r", "--contains", self.commit],
                            chdir=self.sourcedir)[1]
        for branch in branches.decode('utf-8').split("\n"):
            branch = branch.strip()
            if branch[len("origin/"):] == self.branch:
                found = True
                break
        if not found:
            raise ValidationError("Commit %s is not in branch %s." % (self.commit, self.branch))

    def scm_url_from_name(self, name):
        """
        Generates new SCM URL for another module defined by a name. The new URL
        is based on the root of current SCM URL.
        """
        if self.scheme == "git":
            return self.repository_root + name + ".git"

        return None

    @staticmethod
    @retry(
        timeout=conf.scm_net_timeout,
        interval=conf.scm_net_retry_interval,
        wait_on=UnprocessableEntity)
    def _run(cmd, chdir=None, log_stdout=False):
        proc = sp.Popen(cmd, stdout=sp.PIPE, stderr=sp.PIPE, cwd=chdir)
        stdout, stderr = proc.communicate()
        if log_stdout and stdout:
            log.debug(stdout)
        if stderr:
            log.warning(stderr)
        if proc.returncode != 0:
            raise UnprocessableEntity("Failed on %r, retcode %r, out %r, err %r" % (
                cmd, proc.returncode, stdout, stderr))
        return proc.returncode, stdout, stderr

    def checkout(self, scmdir):
        """Checkout the module from SCM.

        :param str scmdir: The working directory
        :returns: str -- the directory that the module was checked-out into
        :raises: RuntimeError
        """
        # TODO: sanity check arguments
        if self.scheme == "git":
            self.sourcedir = '%s/%s' % (scmdir, self.name)

            module_clone_cmd = ['git', 'clone', '-q']
            if self.commit:
                module_checkout_cmd = ['git', 'checkout', '-q', self.commit]
            else:
                module_clone_cmd.extend(['--depth', '1'])
            module_clone_cmd.extend([self.repository, self.sourcedir])

            # perform checkouts
            SCM._run(module_clone_cmd, chdir=scmdir)
            if self.commit:
                try:
                    SCM._run(module_checkout_cmd, chdir=self.sourcedir)
                except RuntimeError as e:
                    if (e.message.endswith(
                            " did not match any file(s) known to git.\\n\"") or
                       "fatal: reference is not a tree: " in e.message):
                        raise UnprocessableEntity(
                            "checkout: The requested commit hash was not found "
                            "within the repository. Perhaps you forgot to push. "
                            "The original message was: %s" % e.message)
                    raise

            timestamp = SCM._run(["git", "show", "-s", "--format=%ct"], chdir=self.sourcedir)[1]
            dt = datetime.datetime.utcfromtimestamp(int(timestamp))
            self.version = dt.strftime("%Y%m%d%H%M%S")
        else:
            raise RuntimeError("checkout: Unhandled SCM scheme.")
        return self.sourcedir

    def get_latest(self, ref='master'):
        """ Get the latest commit hash based on the provided git ref

        :param ref: a string of a git ref (either a branch or commit hash)
        :returns: a string of the latest commit hash
        :raises: RuntimeError
        """
        if ref is None:
            ref = 'master'
        if self.scheme == "git":
            log.debug("Getting/verifying commit hash for %s" % self.repository)
            # get all the branches on the remote
            output = SCM._run(["git", "ls-remote", "--exit-code", self.repository])[1]
            # pair branch names and their latest refs into a dict. The output of the above command
            # is multiple lines of "bf028e573e7c18533d89c7873a411de92d4d913e	refs/heads/master".
            # So the dictionary ends up in the format of
            # {"master": "bf028e573e7c18533d89c7873a411de92d4d913e"...}.
            branches = {}
            for branch_and_ref in output.decode('utf-8').strip().split("\n"):
                # Only look at refs/heads/* and not refs/remotes/origin/*
                if "refs/heads/" not in branch_and_ref:
                    continue
                # This grabs the last bit of text after the last "/", which is the branch name
                cur_branch = branch_and_ref.split("\t")[-1].split("refs/heads/")[-1]
                # This grabs the text before the first tab, which is the commit hash
                cur_ref = branch_and_ref.split("\t")[0]
                branches[cur_branch] = cur_ref
            # first check if the branch name is in the repo
            if ref in branches:
                return branches[ref]
            # if the branch is not in the repo it may be a ref.
            else:
                # The call below will either return the commit hash as is (if a full one was
                # provided) or the full commit hash (if a short hash was provided). If ref is not
                # a commit hash, then this will raise an exception.
                return self.get_full_commit_hash(commit_hash=ref)
        else:
            raise RuntimeError("get_latest: Unhandled SCM scheme.")

    def get_full_commit_hash(self, commit_hash=None):
        """
        Takes a shortened commit hash and returns the full hash
        :param commit_hash: a shortened commit hash. If not specified, the
        one in the URL will be used
        :return: string of the full commit hash
        """
        if commit_hash:
            commit_to_check = commit_hash
        elif self.commit:
            commit_to_check = self.commit
        else:
            raise RuntimeError('No commit hash was specified for "{0}"'.format(
                self.url))

        if self.scheme == 'git':
            log.debug('Getting the full commit hash for "{0}"'
                      .format(self.repository))
            td = None
            try:
                td = tempfile.mkdtemp()
                SCM._run(['git', 'clone', '-q', self.repository, td, '--bare'])
                output = SCM._run(
                    ['git', 'rev-parse', commit_to_check], chdir=td)[1]
            finally:
                if td and os.path.exists(td):
                    shutil.rmtree(td)

            if output:
                return str(output.decode('utf-8').strip('\n'))

            raise UnprocessableEntity(
                'The full commit hash of "{0}" for "{1}" could not be found'
                .format(commit_hash, self.repository))
        else:
            raise RuntimeError('get_full_commit_hash: Unhandled SCM scheme.')

    def get_module_yaml(self):
        """
        Get full path to the module's YAML file.

        :return: path as a string
        :raises UnprocessableEntity
        """
        if not self.sourcedir:
            raise ProgrammingError("Do .checkout() first.")

        path_to_yaml = os.path.join(self.sourcedir, (self.name + ".yaml"))
        try:
            with open(path_to_yaml):
                return path_to_yaml
        except IOError:
            log.error("get_module_yaml: The SCM repository doesn't contain a modulemd file. "
                      "Couldn't access: %s" % path_to_yaml)
            raise UnprocessableEntity("The SCM repository doesn't contain a modulemd file")

    @staticmethod
    def is_full_commit_hash(scheme, commit):
        """
        Determines if a commit hash is the full commit hash. For instance, if
        the scheme is git, it will determine if the commit is a full SHA1 hash
        :param scheme: a string containing the SCM type (e.g. git)
        :param commit: a string containing the commit
        :return: boolean
        """
        if scheme == 'git':
            sha1_pattern = re.compile(r'^[0-9a-f]{40}$')
            return bool(re.match(sha1_pattern, commit))
        else:
            raise RuntimeError('is_full_commit_hash: Unhandled SCM scheme.')

    def is_available(self, strict=False):
        """Check whether the scmurl is available for checkout.

        :param bool strict: When True, raise expection on error instead of
                            returning False.
        :returns: bool -- the scmurl is available for checkout
        """
        td = None
        try:
            td = tempfile.mkdtemp()
            self.checkout(td)
            return True
        except Exception:
            if strict:
                raise
            return False
        finally:
            try:
                if td is not None:
                    shutil.rmtree(td)
            except Exception as e:
                log.warning(
                    "Failed to remove temporary directory {!r}: {}".format(
                        td, str(e)))

    @property
    def url(self):
        """The original scmurl."""
        return self._url

    @url.setter
    def url(self, s):
        self._url = str(s)

    @property
    def scheme(self):
        """The SCM scheme."""
        return self._scheme

    @scheme.setter
    def scheme(self, s):
        self._scheme = str(s)

    @property
    def sourcedir(self):
        """The SCM source directory."""
        return self._sourcedir

    @sourcedir.setter
    def sourcedir(self, s):
        self._sourcedir = str(s)

    @property
    def repository(self):
        """The repository part of the scmurl."""
        return self._repository

    @repository.setter
    def repository(self, s):
        self._repository = str(s)

    @property
    def commit(self):
        """The commit ID, for example the git hash, or None."""
        return self._commit

    @commit.setter
    def commit(self, s):
        self._commit = str(s) if s else None

    @property
    def name(self):
        """The module name."""
        return self._name

    @name.setter
    def name(self, s):
        self._name = str(s)



"""Remove the COPR columns

Revision ID: 9d5e6938588f
Revises: 708ac8950f55
Create Date: 2018-06-28 13:57:08.977877

"""

revision = '9d5e6938588f'
down_revision = '708ac8950f55'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.drop_column('module_builds', 'copr_project')
    op.drop_column('module_builds', 'copr_owner')


def downgrade():
    op.add_column('module_builds', sa.Column('copr_owner', sa.VARCHAR(), nullable=True))
    op.add_column('module_builds', sa.Column('copr_project', sa.VARCHAR(), nullable=True))



# -*- coding: utf-8 -*-


# Copyright (c) 2018  Red Hat, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# Written by Martin Curlej <mcurlej@redhat.com>

"""MBS handler functions."""

import logging
import kobo.rpmlib
import requests

from module_build_service import db
from module_build_service import models
from module_build_service.errors import UnprocessableEntity
from module_build_service.resolver.base import GenericResolver

log = logging.getLogger()


class MBSResolver(GenericResolver):

    backend = "mbs"

    def __init__(self, config):
        self.mbs_prod_url = config.mbs_url
        self.session = requests.Session()
        adapter = requests.adapters.HTTPAdapter(max_retries=3)
        self.session.mount(self.mbs_prod_url, adapter)

    def _query_from_nsvc(self, name, stream, version=None, context=None, state="ready"):
        """
        Generates dict with MBS params query.

        :param str name: Name of the module to query.
        :param str stream: Stream of the module to query.
        :param str version/int: Version of the module to query.
        :param str context: Context of the module to query.
        """
        query = {
            "name": name,
            "stream": stream,
            "state": state,
            "verbose": True,
            "order_desc_by": "version"
        }
        if version is not None:
            query["version"] = str(version)
        if context is not None:
            query["context"] = context
        return query

    def _get_modules(self, name, stream, version=None, context=None, state="ready", strict=False):
        """
        :param module_info: str, mmd or module dict
        :param strict: Normally this function returns None if no module can be
               found.  If strict=True, then an UnprocessableEntity is raised.

        :return final list of module_info which pass repoclosure
        """
        query = self._query_from_nsvc(name, stream, version, context, state)
        query["page"] = 1
        query["per_page"] = 10
        modules = []

        while True:
            res = self.session.get(self.mbs_prod_url, params=query)
            if not res.ok:
                raise RuntimeError("Failed to query MBS with query %r returned HTTP status %s" %
                                   (query, res.status_code))
                break

            data = res.json()
            modules_per_page = data["items"]
            modules += modules_per_page

            if not data["meta"]["next"]:
                break

            query["page"] += 1

        # Error handling
        if not modules:
            if strict:
                raise UnprocessableEntity("Failed to find module in MBS %r" % query)
            else:
                return None

        return modules

    def _get_module(self, name, stream, version, context, state="ready", strict=False):
        return self._get_modules(name, stream, version, context, state, strict)[0]

    def get_module_modulemds(self, name, stream, version=None, context=None, strict=False):
        """
        Gets the module modulemds from the resolver.
        :param name: a string of the module's name
        :param stream: a string of the module's stream
        :param version: a string or int of the module's version. When None, latest version will
            be returned.
        :param context: a string of the module's context. When None, all contexts will
            be returned.
        :kwarg strict: Normally this function returns [] if no module can be
            found.  If strict=True, then a UnprocessableEntity is raised.
        :return: List of Modulemd metadata instances matching the query
        """
        yaml = None

        local_modules = models.ModuleBuild.local_modules(db.session, name, stream)
        if local_modules:
            return [m.mmd() for m in local_modules]

        modules = self._get_modules(name, stream, version, context, strict=strict)
        if not modules:
            return []

        mmds = []
        for module in modules:
            if module:
                yaml = module['modulemd']

            if not yaml:
                if strict:
                    raise UnprocessableEntity(
                        "Failed to find modulemd entry in MBS for %r" % module)
                else:
                    return None

            mmds.append(self.extract_modulemd(yaml, strict=strict))
        return mmds

    def resolve_profiles(self, mmd, keys):
        """
        :param mmd: ModuleMetadata instance of module
        :param keys: list of modulemd installation profiles to include in
                     the result.
        :return: Dictionary with keys set according to `keys` param and values
                 set to union of all components defined in all installation
                 profiles matching the key using the buildrequires.

        If there are some modules loaded by utils.load_local_builds(...), these
        local modules will be considered when returning the profiles.

        https://pagure.io/fm-orchestrator/issue/181
        """

        results = {}
        for key in keys:
            results[key] = set()
        for module_name, module_info in mmd.get_xmd()['mbs']['buildrequires'].items():
            local_modules = models.ModuleBuild.local_modules(
                db.session, module_name, module_info['stream'])
            if local_modules:
                local_module = local_modules[0]
                log.info("Using local module %r to resolve profiles.",
                         local_module)
                dep_mmd = local_module.mmd()
                for key in keys:
                    if key in dep_mmd.get_profiles().keys():
                        results[key] |= set(dep_mmd.get_profiles()[key].get_rpms().get())
                continue

            # Find the dep in the built modules in MBS
            modules = self._get_modules(
                module_name, module_info['stream'], module_info['version'],
                module_info['context'], strict=True)

            for module in modules:
                yaml = module['modulemd']
                dep_mmd = self.extract_modulemd(yaml)
                # Take note of what rpms are in this dep's profile.
                for key in keys:
                    if key in dep_mmd.get_profiles().keys():
                        results[key] |= set(dep_mmd.get_profiles()[key].get_rpms().get())

        # Return the union of all rpms in all profiles of the given keys.
        return results

    def get_module_build_dependencies(self, name=None, stream=None, version=None, context=None,
                                      mmd=None, strict=False):
        """
        :param name: a module's name (required if mmd is not set)
        :param stream: a module's stream (required if mmd is not set)
        :param version: a module's version (required if mmd is not set)
        :param context: a module's context (required if mmd is not set)
        :param mmd: uses the mmd instead of the name, stream, version
        :param strict: Normally this function returns None if no module can be
            found.  If strict=True, then an UnprocessableEntity is raised.
        :return dict with koji_tag as a key and ModuleMetadata object as value.
        """

        if mmd:
            log.debug("get_module_build_dependencies(mmd=%r strict=%r)" % (mmd, strict))
        elif any(x is None for x in [name, stream, version, context]):
            raise RuntimeError('The name, stream, version, and/or context weren\'t specified')
        else:
            version = str(version)
            log.debug("get_module_build_dependencies(%s, strict=%r)"
                      % (', '.join([name, stream, str(version), context]), strict))

        # This is the set we're going to build up and return.
        module_tags = {}

        if mmd:
            queried_mmd = mmd
        else:
            queried_module = self._get_module(
                name, stream, version, context, strict=strict)
            yaml = queried_module['modulemd']
            queried_mmd = self.extract_modulemd(yaml, strict=strict)

        if (not queried_mmd or not queried_mmd.get_xmd().get('mbs') or
                'buildrequires' not in queried_mmd.get_xmd()['mbs'].keys()):
            raise RuntimeError(
                'The module "{0!r}" did not contain its modulemd or did not have '
                'its xmd attribute filled out in MBS'.format(queried_mmd))

        buildrequires = queried_mmd.get_xmd()['mbs']['buildrequires']
        # Queue up the next tier of deps that we should look at..
        for name, details in buildrequires.items():
            local_modules = models.ModuleBuild.local_modules(
                db.session, name, details['stream'])
            if local_modules:
                for m in local_modules:
                    module_tags[m.koji_tag] = m.mmd()
                continue

            if "context" not in details:
                details["context"] = "00000000"
            modules = self._get_modules(
                name, details['stream'], details['version'],
                details['context'], strict=True)
            for m in modules:
                if m["koji_tag"] in module_tags:
                    continue
                module_tags[m["koji_tag"]] = self.extract_modulemd(m["modulemd"])

        return module_tags

    def resolve_requires(self, requires):
        """
        Takes `requires` dict with module_name as key and module_stream as value.
        Resolves the stream to particular latest version of a module and returns
        new dict in following format:

        {
            "module_name": {
                "ref": module_commit_hash,
                "stream": original_module_stream,
                "version": module_version,
                "filtered_rpms": ["nvr", ...]
            },
            ...
        }

        If there are some modules loaded by utils.load_local_builds(...), these
        local modules will be considered when resolving the requires.

        Raises RuntimeError on MBS lookup error.
        """
        new_requires = {}
        for nsvc in requires:
            nsvc_splitted = nsvc.split(":")
            if len(nsvc_splitted) == 2:
                module_name, module_stream = nsvc_splitted
                module_version = None
                module_context = None
            elif len(nsvc_splitted) == 4:
                module_name, module_stream, module_version, module_context = nsvc_splitted
            else:
                raise ValueError(
                    "Only N:S or N:S:V:C is accepted by resolve_requires, got %s" % nsvc)
            # Try to find out module dependency in the local module builds
            # added by utils.load_local_builds(...).
            local_modules = models.ModuleBuild.local_modules(
                db.session, module_name, module_stream)
            if local_modules:
                local_build = local_modules[0]
                new_requires[module_name] = {
                    # The commit ID isn't currently saved in modules.yaml
                    'ref': None,
                    'stream': local_build.stream,
                    'version': local_build.version,
                    'context': local_build.context,
                    # No need to set filtered_rpms for local builds, because MBS
                    # filters the RPMs automatically when the module build is
                    # done.
                    'filtered_rpms': []
                }
                continue

            commit_hash = None
            version = None
            filtered_rpms = []
            module = self._get_module(
                module_name, module_stream, module_version,
                module_context, strict=True)
            if module.get('modulemd'):
                mmd = self.extract_modulemd(module['modulemd'])
                if mmd.get_xmd().get('mbs') and 'commit' in mmd.get_xmd()['mbs'].keys():
                    commit_hash = mmd.get_xmd()['mbs']['commit']

                # Find out the particular NVR of filtered packages
                if "rpms" in module and mmd.get_rpm_filter().get():
                    for rpm in module["rpms"]:
                        nvr = kobo.rpmlib.parse_nvra(rpm)
                        # If the package is not filtered, continue
                        if not nvr["name"] in mmd.get_rpm_filter().get():
                            continue

                        # If the nvr is already in filtered_rpms, continue
                        nvr = kobo.rpmlib.make_nvr(nvr, force_epoch=True)
                        if nvr in filtered_rpms:
                            continue
                        filtered_rpms.append(nvr)

            if module.get('version'):
                version = module['version']

            if version and commit_hash:
                new_requires[module_name] = {
                    'ref': commit_hash,
                    'stream': module_stream,
                    'version': str(version),
                    'context': module["context"],
                    'filtered_rpms': filtered_rpms,
                }
            else:
                raise RuntimeError(
                    'The module "{0}" didn\'t contain either a commit hash or a'
                    ' version in MBS'.format(module_name))

        return new_requires
